@@ -2,21 +2,24 @@
  * License, v. 2.0. If a copy of the MPL was not distributed with this
  * file, You can obtain one at http://mozilla.org/MPL/2.0/. */
 
-use crate::android::AndroidHandler;
+use crate::android::{AndroidError, AndroidHandler};
 use crate::capabilities::{FirefoxOptions, ProfileType};
 use crate::logging;
 use crate::prefs;
 use mozprofile::preferences::Pref;
 use mozprofile::profile::{PrefFile, Profile};
 use mozrunner::runner::{FirefoxProcess, FirefoxRunner, Runner, RunnerProcess};
+use std::env;
 use std::fs;
-use std::io::Read;
+use std::io::prelude::*;
 use std::path::{Path, PathBuf};
 use std::time;
+use uuid::Uuid;
 use webdriver::error::{ErrorStatus, WebDriverError, WebDriverResult};
 
 /// A running Gecko instance.
 #[derive(Debug)]
+#[allow(clippy::large_enum_variant)]
 pub(crate) enum Browser {
     Local(LocalBrowser),
     Remote(RemoteBrowser),
@@ -55,6 +58,39 @@ impl Browser {
             }
         }
     }
+
+    pub(crate) fn create_file(&self, content: &[u8]) -> WebDriverResult<String> {
+        let addon_file = format!("addon-{}.xpi", Uuid::new_v4());
+        match self {
+            Browser::Remote(x) => {
+                let path = x.push_file(content, &addon_file).map_err(|e| {
+                    WebDriverError::new(
+                        ErrorStatus::UnknownError,
+                        format!("Failed to create an addon file: {}", e),
+                    )
+                })?;
+
+                Ok(path)
+            }
+            Browser::Local(_) | Browser::Existing(_) => {
+                let path = env::temp_dir().as_path().join(addon_file);
+                let mut xpi_file = fs::File::create(&path).map_err(|e| {
+                    WebDriverError::new(
+                        ErrorStatus::UnknownError,
+                        format!("Failed to create an addon file: {}", e),
+                    )
+                })?;
+                xpi_file.write_all(content).map_err(|e| {
+                    WebDriverError::new(
+                        ErrorStatus::UnknownError,
+                        format!("Failed to write data to the addon file: {}", e),
+                    )
+                })?;
+
+                Ok(path.display().to_string())
+            }
+        }
+    }
 }
 
 #[derive(Debug)]
@@ -63,7 +99,7 @@ pub(crate) struct LocalBrowser {
     marionette_port: u16,
     prefs_backup: Option<PrefsBackup>,
     process: FirefoxProcess,
-    profile_path: Option<PathBuf>,
+    pub(crate) profile_path: Option<PathBuf>,
 }
 
 impl LocalBrowser {
@@ -71,6 +107,8 @@ impl LocalBrowser {
         options: FirefoxOptions,
         marionette_port: u16,
         jsdebugger: bool,
+        system_access: bool,
+        profile_root: Option<&Path>,
     ) -> WebDriverResult<LocalBrowser> {
         let binary = options.binary.ok_or_else(|| {
             WebDriverError::new(
@@ -87,7 +125,7 @@ impl LocalBrowser {
         let mut profile = match options.profile {
             ProfileType::Named => None,
             ProfileType::Path(x) => Some(x),
-            ProfileType::Temporary => Some(Profile::new()?),
+            ProfileType::Temporary => Some(Profile::new(profile_root)?),
         };
 
         let (profile_path, prefs_backup) = if let Some(ref mut profile) = profile {
@@ -117,6 +155,9 @@ impl LocalBrowser {
         if jsdebugger {
             runner.arg("--jsdebugger");
         }
+        if system_access {
+            runner.arg("--remote-allow-system-access");
+        }
         if let Some(args) = options.args.as_ref() {
             runner.args(args);
         }
@@ -159,10 +200,12 @@ impl LocalBrowser {
             }
         }
         self.process.kill()?;
+
         // Restoring the prefs if the browser fails to stop perhaps doesn't work anyway
         if let Some(prefs_backup) = self.prefs_backup {
             prefs_backup.restore();
         };
+
         Ok(())
     }
 
@@ -225,8 +268,9 @@ fn read_marionette_port(profile_path: &Path) -> Option<u16> {
 #[derive(Debug)]
 /// A remote instance, running on a (target) Android device.
 pub(crate) struct RemoteBrowser {
-    handler: AndroidHandler,
+    pub(crate) handler: AndroidHandler,
     marionette_port: u16,
+    prefs_backup: Option<PrefsBackup>,
 }
 
 impl RemoteBrowser {
@@ -234,10 +278,12 @@ impl RemoteBrowser {
         options: FirefoxOptions,
         marionette_port: u16,
         websocket_port: Option<u16>,
+        system_access: bool,
+        profile_root: Option<&Path>,
     ) -> WebDriverResult<RemoteBrowser> {
         let android_options = options.android.unwrap();
 
-        let handler = AndroidHandler::new(&android_options, marionette_port, websocket_port)?;
+        let handler = AndroidHandler::new(&android_options, marionette_port, system_access, websocket_port)?;
 
         // Profile management.
         let (mut profile, is_custom_profile) = match options.profile {
@@ -248,10 +294,10 @@ impl RemoteBrowser {
                 ));
             }
             ProfileType::Path(x) => (x, true),
-            ProfileType::Temporary => (Profile::new()?, false),
+            ProfileType::Temporary => (Profile::new(profile_root)?, false),
         };
 
-        set_prefs(
+        let prefs_backup = set_prefs(
             handler.marionette_target_port,
             &mut profile,
             is_custom_profile,
@@ -272,11 +318,18 @@ impl RemoteBrowser {
         Ok(RemoteBrowser {
             handler,
             marionette_port,
+            prefs_backup,
         })
     }
 
     fn close(self) -> WebDriverResult<()> {
         self.handler.force_stop()?;
+
+        // Restoring the prefs if the browser fails to stop perhaps doesn't work anyway
+        if let Some(prefs_backup) = self.prefs_backup {
+            prefs_backup.restore();
+        };
+
         Ok(())
     }
 
@@ -287,6 +340,10 @@ impl RemoteBrowser {
     fn update_marionette_port(&mut self, port: u16) {
         self.marionette_port = port;
     }
+
+    fn push_file(&self, content: &[u8], path: &str) -> Result<String, AndroidError> {
+        self.handler.push_as_file(content, path)
+    }
 }
 
 fn set_prefs(
@@ -378,6 +435,8 @@ mod tests {
     use super::set_prefs;
     use crate::browser::read_marionette_port;
     use crate::capabilities::{FirefoxOptions, ProfileType};
+    use base64::prelude::BASE64_STANDARD;
+    use base64::Engine;
     use mozprofile::preferences::{Pref, PrefValue};
     use mozprofile::profile::Profile;
     use serde_json::{Map, Value};
@@ -390,7 +449,7 @@ mod tests {
         let mut profile_data = Vec::with_capacity(1024);
         let mut profile = File::open("src/tests/profile.zip").unwrap();
         profile.read_to_end(&mut profile_data).unwrap();
-        Value::String(base64::encode(&profile_data))
+        Value::String(BASE64_STANDARD.encode(&profile_data))
     }
 
     // This is not a pretty test, mostly due to the nature of
@@ -398,7 +457,7 @@ mod tests {
     // several regressions related to remote.log.level.
     #[test]
     fn test_remote_log_level() {
-        let mut profile = Profile::new().unwrap();
+        let mut profile = Profile::new(None).unwrap();
         set_prefs(2828, &mut profile, false, vec![], false).ok();
         let user_prefs = profile.user_prefs().unwrap();
 
@@ -460,7 +519,7 @@ mod tests {
 
     #[test]
     fn test_pref_backup() {
-        let mut profile = Profile::new().unwrap();
+        let mut profile = Profile::new(None).unwrap();
 
         // Create some prefs in the profile
         let initial_prefs = profile.user_prefs().unwrap();
@@ -470,7 +529,7 @@ mod tests {
         let prefs_path = initial_prefs.path.clone();
 
         let mut conflicting_backup_path = initial_prefs.path.clone();
-        conflicting_backup_path.set_extension("geckodriver_backup".to_string());
+        conflicting_backup_path.set_extension("geckodriver_backup");
         println!("{:?}", conflicting_backup_path);
         let mut file = File::create(&conflicting_backup_path).unwrap();
         file.write_all(b"test").unwrap();
@@ -489,7 +548,7 @@ mod tests {
 
         assert!(user_prefs.path.exists());
         let mut backup_path = user_prefs.path.clone();
-        backup_path.set_extension("geckodriver_backup_1".to_string());
+        backup_path.set_extension("geckodriver_backup_1");
 
         assert!(backup_path.exists());
 
